#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2010, 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#

#
# This test module contains misc compound tests which do not have
# their own module yet.
#

# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0

# global utility functions
compound bracketstat=(
	integer bopen=0
	integer bclose=0
)

function count_brackets
{
	typeset x="$1"
	typeset c

	integer i
	(( bracketstat.bopen=0 , bracketstat.bclose=0 ))

	for (( i=0 ; i < ${#x} ; i++ )) ; do
	        c="${x:i:1}"
		[[ "$c" == '(' ]] && (( bracketstat.bopen++ ))
		[[ "$c" == ')' ]] && (( bracketstat.bclose++ ))
	done
	
	(( bracketstat.bopen != bracketstat.bclose )) && return 1
	
	return 0
}


typeset ocwd
typeset tmpdir

# create temporary test directory
ocwd="$PWD"
tmpdir="$(mktemp -t -d "test_sun_solaris_compound_misc.XXXXXXXX")" || err_exit "Cannot create temporary directory"

cd "${tmpdir}" || { err_exit "cd ${tmpdir} failed." ; exit $((Errors<125?Errors:125)) ; }

# ksh93 <= ast-ksh.2010-03-09 prints garbage for compound x=( compound -a nodes=( [4]=( ) ) );typeset -p x
function test_compound_indexed_array_init_1
{
	compound vx=( compound -a nodes=( [4]=( ) )  )
	compound vy
	compound -a vy.nodes=( [4]=( ) )
	compound vz
	compound -a vz.nodes
	vz.nodes[4]=( )

	cx="$(typeset -p vx)" ; cx="${cx//vx/tt}"
	cy="$(typeset -p vy)" ; cy="${cy//vy/tt}"
	cz="$(typeset -p vz)" ; cz="${cz//vz/tt}"
	[[ "$cx" == "$cy" ]] || err_exit "'$cx' != '$cy'"
	[[ "$cx" == "$cz" ]] || err_exit "'$cx' != '$cz'"
	[[ "$cy" == "$cz" ]] || err_exit "'$cy' != '$cz'"

	count_brackets "$cx" || err_exit "Brackets not balanced for '$cx'"
	count_brackets "$cy" || err_exit "Brackets not balanced for '$cy'"
	count_brackets "$cz" || err_exit "Brackets not balanced for '$cz'"
	count_brackets "$(print -v vx)" || err_exit "Brackets not balanced for '$(print -v vx)'"
	count_brackets "$(print -v vy)" || err_exit "Brackets not balanced for '$(print -v vy)'"
	count_brackets "$(print -v vz)" || err_exit "Brackets not balanced for '$(print -v vz)'"
	count_brackets "$(print -C vx)" || err_exit "Brackets not balanced for '$(print -C vx)'"
	count_brackets "$(print -C vy)" || err_exit "Brackets not balanced for '$(print -C vy)'"
	count_brackets "$(print -C vz)" || err_exit "Brackets not balanced for '$(print -C vz)'"

	cx="$(typeset +p vx.nodes)" ;    [[ "$cx" == *-C* && "$cx" == *-a* ]] || err_exit "'$cx' lacks -C/-a attribute"
	cy="$(typeset +p vy.nodes)" ;    [[ "$cy" == *-C* && "$cy" == *-a* ]] || err_exit "'$cy' lacks -C/-a attribute"
	cz="$(typeset +p vz.nodes)" ;    [[ "$cz" == *-C* && "$cz" == *-a* ]] || err_exit "'$cz' lacks -C/-a attribute"
	cx="$(typeset +p vx.nodes[4])" ; [[ "$cx" == *-C*                  ]] || err_exit "'$cx' lacks -C attribute"
	cy="$(typeset +p vy.nodes[4])" ; [[ "$cy" == *-C*                  ]] || err_exit "'$cy' lacks -C attribute"
	cz="$(typeset +p vz.nodes[4])" ; [[ "$cz" == *-C*                  ]] || err_exit "'$cz' lacks -C attribute"

	return 0
}

# ksh93 <= ast-ksh.2010-03-09 prints garbage for compound x=( compound -a nodes=( [4]=( ) ) );typeset -p x
# this test is the same as test_compound_indexed_array_init_1 but "-a" was replaced with "-A"
function test_compound_associative_array_init_1
{
	compound vx=( compound -A nodes=( [4]=( ) )  )
	compound vy
	compound -A vy.nodes=( [4]=( ) )
	compound vz
	compound -A vz.nodes
	vz.nodes[4]=( )

	cx="$(typeset -p vx)" ; cx="${cx//vx/tt}"
	cy="$(typeset -p vy)" ; cy="${cy//vy/tt}"
	cz="$(typeset -p vz)" ; cz="${cz//vz/tt}"
	[[ "$cx" == "$cy" ]] || err_exit "'$cx' != '$cy'"
	[[ "$cx" == "$cz" ]] || err_exit "'$cx' != '$cz'"
	[[ "$cy" == "$cz" ]] || err_exit "'$cy' != '$cz'"

	count_brackets "$cx" || err_exit "Brackets not balanced for '$cx'"
	count_brackets "$cy" || err_exit "Brackets not balanced for '$cy'"
	count_brackets "$cz" || err_exit "Brackets not balanced for '$cz'"
	count_brackets "$(print -v vx)" || err_exit "Brackets not balanced for '$(print -v vx)'"
	count_brackets "$(print -v vy)" || err_exit "Brackets not balanced for '$(print -v vy)'"
	count_brackets "$(print -v vz)" || err_exit "Brackets not balanced for '$(print -v vz)'"
	count_brackets "$(print -C vx)" || err_exit "Brackets not balanced for '$(print -C vx)'"
	count_brackets "$(print -C vy)" || err_exit "Brackets not balanced for '$(print -C vy)'"
	count_brackets "$(print -C vz)" || err_exit "Brackets not balanced for '$(print -C vz)'"

	cx="$(typeset +p vx.nodes)" ;    [[ "$cx" == *-C* && "$cx" == *-A* ]] || err_exit "'$cx' lacks -C/-A attribute"
	cy="$(typeset +p vy.nodes)" ;    [[ "$cy" == *-C* && "$cy" == *-A* ]] || err_exit "'$cy' lacks -C/-A attribute"
	cz="$(typeset +p vz.nodes)" ;    [[ "$cz" == *-C* && "$cz" == *-A* ]] || err_exit "'$cz' lacks -C/-A attribute"
	cx="$(typeset +p vx.nodes[4])" ; [[ "$cx" == *-C*                  ]] || err_exit "'$cx' lacks -C attribute"
	cy="$(typeset +p vy.nodes[4])" ; [[ "$cy" == *-C*                  ]] || err_exit "'$cy' lacks -C attribute"
	cz="$(typeset +p vz.nodes[4])" ; [[ "$cz" == *-C*                  ]] || err_exit "'$cz' lacks -C attribute"

	return 0
}

# test whether the initalisatio of arrays in a compound variable works
function test_array_in_compound_init
{
        compound out=( typeset stdout stderr ; integer res )
	typeset testname
	integer i
	compound -r -a tests=(
		( cmd='compound c ; compound -a c.c2 ; nameref x=c.c2[4].o ; integer -a x ; typeset +p c.c2[4].o' )
		( cmd='compound c ; compound -a c.c2 ; nameref x=c.c2[4].o ; integer -A x ; typeset +p c.c2[4].o' )
		( cmd='compound c ; compound -A c.c2 ; nameref x=c.c2[4].o ; integer -a x ; typeset +p c.c2[4].o' )
		( cmd='compound c ; compound -A c.c2 ; nameref x=c.c2[4].o ; integer -A x ; typeset +p c.c2[4].o' )
	)

	for (( i=0 ; i < ${#tests[@]} ; i++ )) ; do
		nameref tst=tests[i]
		testname="${0}/${tst.cmd}"

		out.stderr="${ { out.stdout="${ ${SHELL} -o nounset -o errexit -c "${tst.cmd}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"

	        [[ "${out.stdout}" != *-C*      ]] || err_exit "${testname}: Expected stdout to not to match *-C*, got $(printf '%q\n' "${out.stdout}")"
       		[[ "${out.stderr}" == ''	]] || err_exit "${testname}: Expected empty stderr, got $(printf '%q\n' "${out.stderr}")"
		(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"
	done
	
	return 0
}


# test whether the initalisatio of arrays in a compound variable works
function test_array_append_through_nameref
{
        compound out=( typeset stdout stderr ; integer res )
	typeset testname
	integer i
	compound -r -a tests=(
		( cmd='compound c ; integer -a c.x ;                 c.x+=( 4 ) ; c.x+=( 2 ) ; print -v c' )
		( cmd='compound c=( integer -a x ) ;                 c.x+=( 4 ) ; c.x+=( 2 ) ; print -v c' )
		( cmd='compound c ; integer -a c.x ; nameref l=c.x ; l+=( 4 ) ;   l+=( 2 ) ;   print -v c' )
		( cmd='compound c=( integer -a x ) ; nameref l=c.x ; l+=( 4 ) ;   l+=( 2 ) ;   print -v c' )
	)

	for (( i=0 ; i < ${#tests[@]} ; i++ )) ; do
		nameref tst=tests[i]
		testname="${0}/${tst.cmd}"

		out.stderr="${ { out.stdout="${ ${SHELL} -o nounset -o errexit -c "${tst.cmd}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"

	        [[ "${out.stdout}" != *-C*      ]] || err_exit "${testname}: Expected stdout to not to match *-C*, got $(printf '%q\n' "${out.stdout}")"
       		[[ "${out.stderr}" == ''	]] || err_exit "${testname}: Expected empty stderr, got $(printf '%q\n' "${out.stderr}")"
		(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"
	done
	
	return 0
}

# test whether creating an empty compound variable through a nameref in
# an accociative array works
# ast-ksh.2010-09-24 was printing "parent not found".
function test_implicit_node_creation_through_nameref_in_local_associative_array_in_function
{
        compound out=( typeset stdout stderr ; integer res )
	typeset testname
	integer i
	compound -r -a tests=(
		# this was the version which failed in ast-ksh.2010-09-24 with "no parent"
		( cmd='function bf { nameref treename=$1 ; nodepath="treename" ; nameref x="$nodepath" ; compound -A x.nodes ; nameref node=treename.nodes[4] ; node=() ; typeset +p node.elements ; } ; compound c ; bf c ; print "MARK"' )
		# variant 1: function "bf" declared with "bf()" instead of "function bf"
		( cmd='bf()        { nameref treename=$1 ; nodepath="treename" ; nameref x="$nodepath" ; compound -A x.nodes ; nameref node=treename.nodes[4] ; node=() ; typeset +p node.elements ; } ; compound c ; bf c ; print "MARK"' )
		# like the first version, but "c" is a local variable
		( cmd='function bf { nameref treename=$1 ; nodepath="treename" ; nameref x="$nodepath" ; compound -A x.nodes ; nameref node=treename.nodes[4] ; node=() ; typeset +p node.elements ; } ; function wrap { compound c ; bf c ; } ; wrap ; print "MARK"' )
		# like the 2nd version, but "c" is a local variable
		( cmd='bf()        { nameref treename=$1 ; nodepath="treename" ; nameref x="$nodepath" ; compound -A x.nodes ; nameref node=treename.nodes[4] ; node=() ; typeset +p node.elements ; } ; function wrap { compound c ; bf c ; } ; wrap ; print "MARK"' )
	)

	for (( i=0 ; i < ${#tests[@]} ; i++ )) ; do
		nameref tst=tests[i]
		testname="${0}/${i}"

		out.stderr="${ { out.stdout="${ ${SHELL} -o nounset -o errexit -c "${tst.cmd}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"

	        [[ "${out.stdout}" == 'MARK' ]] || err_exit "${testname}: Expected stdout=='MARK', got $(printf '%q\n' "${out.stdout}")"
       		[[ "${out.stderr}" == ''     ]] || err_exit "${testname}: Expected empty stderr, got $(printf '%q\n' "${out.stderr}")"
		(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"
	done
	
	return 0
}

# run tests
test_compound_indexed_array_init_1
test_compound_associative_array_init_1
test_array_in_compound_init
test_array_append_through_nameref
test_implicit_node_creation_through_nameref_in_local_associative_array_in_function

cd "${ocwd}"
rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."

# tests done
exit $((Errors<125?Errors:125))
